!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   qa_option_mod
!> @brief   module for analysis options
!! @authors Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module qa_option_mod

  use qa_option_str_mod
  use measure_mod
  use select_mod
  use select_atoms_mod
  use molecules_str_mod
  use fileio_grotop_mod
  use fileio_control_mod
  use string_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_opt_info
    logical                         :: check_only     = .true.
  end type s_opt_info

  ! subroutines
  public  :: show_ctrl_option
  public  :: read_ctrl_option
  public  :: setup_option

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_option
  !> @brief        show control parameters in OPTION section
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_option


    write(MsgOut,'(A)') '[SELECTION]'
    write(MsgOut,'(A)') '# group1         = resno:1         # atom group 1'
    write(MsgOut,'(A)') '# group2         = resno:3         # atom group 2'
    write(MsgOut,'(A)') '# group3         = resno:5         # atom group 3'
    write(MsgOut,'(A)') ''
    write(MsgOut,'(A)') '[OPTION]'
    write(MsgOut,'(A)') 'check_only     = YES             # (YES/NO)'
    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_option
  !> @brief        read control parameters in OPTION section
  !! @authors      NT
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   opt_info : OPTION section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_option(handle, opt_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Option'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_opt_info),        intent(inout) :: opt_info


    ! read parameters
    !
    call begin_ctrlfile_section(handle, Section)

    call read_ctrlfile_logical(handle, Section, &
                              'check_only', opt_info%check_only)

    call end_ctrlfile_section(handle)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Option> Parameters of Options'
    if (opt_info%check_only) then
      write(MsgOut,'(A20,A3)') '  check only      = ', 'yes'
    else
      write(MsgOut,'(A20,A2)') '  check only      = ', 'no'
    end if

    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_option
  !> @brief        setup option information
  !! @authors      NT
  !! @param[in]    opt_info : OPTION section control parameters information
  !! @param[in]    sel_info : SELECTION section control parameters information
  !! @param[in]    molecule : molecule information
  !! @param[in]    grotop   : GROTOP information
  !! @param[inout] option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_option(opt_info, sel_info, molecule, grotop, option)

    ! formal arguments
    type(s_opt_info),        intent(in)    :: opt_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_grotop),  target, intent(in)    :: grotop
    type(s_option),  target, intent(inout) :: option

    ! local variables
    integer                  :: i1, i2, imin, imax, alloc_stat
    integer                  :: ncon, ngrp, nspa, nsel
    integer                  :: icon, igrp, ispa, isel

    type(s_grotop_mol), pointer :: gromol
    integer,            pointer :: selidx(:)


    write(MsgOut,'(a)') 'Setup_Option> :'
    write(MsgOut,'(a)') ''

    gromol => grotop%moltypes(1)%mol

    ! check only
    !
    option%check_only = opt_info%check_only


    ! contact list
    !
    ncon = gromol%num_pairs

    write(MsgOut,'(A20,I10)')   '  num contacts    = ', ncon

    allocate(option%contact_list(2,ncon), &
             option%contact_kai   (ncon), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    do icon = 1, ncon

      i1 = gromol%pairs(icon)%atom_idx1
      i2 = gromol%pairs(icon)%atom_idx2

      option%contact_list(1,icon) = i1
      option%contact_list(2,icon) = i2

      option%contact_kai(icon) = compute_dis(molecule%atom_coord(:,i1), &
                                             molecule%atom_coord(:,i2))

    end do


    ! atom selection
    !
    ngrp = size(sel_info%groups)

    write(MsgOut,'(A20,I10)')   '  num sel. groups = ', ngrp

    allocate(option%selatoms(1:ngrp))

    do igrp = 1, ngrp
      call select_atom(molecule, sel_info%groups(igrp), option%selatoms(igrp))

      write(MsgOut,'(a,i12,a,a)') '  group: ', igrp, ' :', &
                                trim(sel_info%groups(igrp))
      write(MsgOut,'(a,i12)')     '    # of selected atom = ', &
                                  size(option%selatoms(igrp)%idx)
      write(MsgOut,'(a)')       ''
    end do


    ! contact selection pair
    !
    nspa = ngrp*(ngrp+1)/2

    write(MsgOut,'(A20,I10)')   '  num sel. pairs  = ', nspa

    allocate(option%contact_sel_pair (ncon), &
             option%sel_pair_nfound  (nspa), &
             option%sel_pair_ncontact(nspa), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    option%contact_sel_pair (1:ncon) = 0
    option%sel_pair_nfound  (1:nspa) = 0
    option%sel_pair_ncontact(1:nspa) = 0

    do icon = 1, ncon

      i1 = 0
      i2 = 0

      do igrp = 1, ngrp

        selidx => option%selatoms(igrp)%idx

        nsel = size(selidx)
        do isel = 1, nsel
          
          if (option%contact_list(1,icon) == selidx(isel)) then
            i1 = igrp
          end if

          if (option%contact_list(2,icon) == selidx(isel)) then
            i2 = igrp
          end if

        end do

        if (i1 /= 0 .and. i2 /= 0) &
          exit

      end do

      if (i1 /= 0 .and. i2 /= 0) then
        imin = min(i1, i2)
        imax = max(i1, i2)
        ispa = (imin - 1) * (2 * ngrp - imin) / 2 + imax
        option%sel_pair_ncontact(ispa) = option%sel_pair_ncontact(ispa) + 1
        option%contact_sel_pair (icon) = ispa

      end if

    end do


    ! q-native value
    !
    allocate(option%qnative(1+nspa), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc


    write(MsgOut,'(a)') ''

    return

  end subroutine setup_option

end module qa_option_mod
