!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_experiments_mod
!> @brief   experimental data
!! @authors Osamu Miyashita (OM), Takaharu Mori (TM)
!
!  (c) Copyright 2016 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
  
module at_experiments_mod

  use molecules_str_mod
  use at_restraints_str_mod
  use at_experiments_str_mod
  use at_enefunc_str_mod
  use fileio_sit_mod
  use fileio_control_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use string_mod
  use timers_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! structures
  type, public :: s_exp_info
    logical                         :: emfit           = .false.
    character(MaxFilename)          :: emfit_target    = ''
    real(wp)                        :: emfit_sigma     = 2.5_wp
    real(wp)                        :: emfit_tolerance = 0.001_wp
    integer                         :: emfit_period    = 1
  end type s_exp_info

  type(s_experiments), target, save :: experiments
  integer,                     save :: natom_local, emfit_icycle
  real(wp),                    save :: corrcoeff_save
  integer,        allocatable, save :: list_local(:)
  integer,        allocatable, save :: icount(:), domain_index(:)
  integer,        allocatable, save :: ig_min_local(:,:), ig_max_local(:,:)
  integer,        allocatable, save :: num_atoms_local(:)
  integer,        allocatable, save :: ig_lower(:,:), ig_upper(:,:)
  real(wp),       allocatable, save :: dot_exp_drhodx(:), dot_exp_drhody(:), dot_exp_drhodz(:)
  real(wp),       allocatable, save :: erfa(:,:), expa(:,:)
  real(wp),       allocatable, save :: derfa_x(:,:), derfa_y(:,:), derfa_z(:,:)
  real(wp),       allocatable, save :: dexpa_x(:,:), dexpa_y(:,:), dexpa_z(:,:)
  real(wp),       allocatable, save :: dot_sim_drhodx(:), dot_sim_drhody(:), dot_sim_drhodz(:)
  real(wp),       allocatable, save :: emfit_force(:,:)

  ! subroutines
  public  :: show_ctrl_experiments
  public  :: read_ctrl_experiments
  public  :: setup_experiments
  private :: setup_experiments_emfit
  public  :: compute_energy_experimental_restraint
  private :: compute_energy_experimental_restraint_emfit

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_experiments
  !> @brief        show EXPERIMENTS section usage
  !! @authors      TM
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "remd", "min", "rpath"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_experiments(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode

    return

  end subroutine show_ctrl_experiments
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      read_ctrl_experiments
  !> @brief        read EXPERIMENTS section in the control file
  !! @authors      TM
  !! @param[in]    handle   : unit number of control files
  !! @param[out]   exp_info : EXPERIMENTS section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_experiments(handle, exp_info) 

    ! parameters
    character(*),            parameter     :: Section = 'Experiments'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_exp_info),        intent(inout) :: exp_info 

    return

  end subroutine read_ctrl_experiments

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      setup_experiments
  !> @brief        setup experiments information
  !! @authors      TM
  !! @param[in]    exp_info    : EXPERIMENTS section control parameters
  !! @param[inout] experiments : experiments information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_experiments(exp_info, molecule, restraints, enefunc)

    ! formal arguments
    type(s_exp_info),        intent(in)    :: exp_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_restraints),      intent(in)    :: restraints
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer    :: i
    logical    :: do_emfit_res

    return

  end subroutine setup_experiments

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      setup_experiments_emfit
  !> @brief        setup experiments information
  !! @authors      OM, TM
  !! @param[in]    exp_info    : EXPERIMENTS section control parameter
  !! @param[inout] experiments : experiments information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_experiments_emfit(exp_info, molecule)

    ! formal arguments
    type(s_exp_info),        intent(in)    :: exp_info
    type(s_molecule),        intent(in)    :: molecule

    ! local variables
    type(s_sit)             :: sit
    integer                 :: i, j, k, n(3), max_nxyz
    real(wp)                :: pi, r, y

    return

  end subroutine setup_experiments_emfit

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_experimental_restraint
  !> @brief        calculate restraint energy from experimental data
  !! @authors      TM
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    coord   : coordinates of target systems
  !! @param[in]    inum    : pointer for restraint function
  !! @param[in]    const   : force constants
  !! @param[in]    ref     : reference value
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial of target systems
  !! @param[out]   eexp    : restraint energy
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_experimental_restraint(enefunc, coord, inum, &
                                           calc_force, force, virial, eexp, cv)

    ! formal arguments
    type(s_enefunc), target, intent(inout) :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: eexp
    real(wp),                intent(inout) :: cv

    return

  end subroutine compute_energy_experimental_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_experimental_restraint_emfit
  !> @brief        calculate restraint energy from experimental data
  !! @authors      OM, TM
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    coord   : coordinates of target systems
  !! @param[in]    inum    : pointer for restraint function
  !! @param[in]    const   : force constants
  !! @param[in]    ref     : reference value
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial of target systems
  !! @param[out]   eexp    : restraint energy
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_experimental_restraint_emfit(enefunc, coord, inum, &
                                     calc_force, force, virial, e_emfit, cv)

    ! formal arguments
    type(s_enefunc), target, intent(inout) :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: e_emfit
    real(wp),                intent(inout) :: cv

    ! local variables
    integer           :: i, j, k, n, m, natom, group_id, ifound
    integer           :: alloc_size, i1, j1, k1
    integer           :: igx_min, igy_min, igz_min, igx_max, igy_max, igz_max
    integer           :: sum_ig, ave_ig
    integer           :: idomain, axis_length(3), iaxis
    integer           :: itmp1, itmp2, icount1
    integer           :: id, omp_get_thread_num
    integer           :: max_natom_local
    real(wp)          :: yzfactor, zfactor, max_length
    real(wp)          :: drho_dx, drho_dy, drho_dz, before_allreduce(2), after_allreduce(2)
    real(wp)          :: dcc_dx, dcc_dy, dcc_dz
    real(wp)          :: norm_sim, dot_exp_sim
    real(wp)          :: f, g, d, coeff_rho, coeff_drho
    real(wp)          :: pi, vol, corrcoeff
    real(wp)          :: dot_exp_drhodx_tmp, dot_exp_drhody_tmp, dot_exp_drhodz_tmp
    real(wp)          :: dot_sim_drhodx_tmp, dot_sim_drhody_tmp, dot_sim_drhodz_tmp
    real(wp)          :: emfit_force_norm, weight
    real(wp)          :: inv_dx, inv_dy, inv_dz
    real(wp)          :: fact1, fact2, fact3
    logical           :: do_allocate
    integer,  pointer :: nx, ny, nz
    integer,  pointer :: n_grid_cut_x, n_grid_cut_y, n_grid_cut_z
    integer,  pointer :: ig(:,:)
    integer,  pointer :: numatoms(:), atom_id(:,:)
    real(wp), pointer :: norm_exp            ! is calculated in setup_emfit just once
    real(wp), pointer :: sigma
    real(wp), pointer :: simulated_map(:,:,:), target_map(:,:,:)
    real(wp), pointer :: bound_x(:), bound_y(:), bound_z(:)

    return

  end subroutine compute_energy_experimental_restraint_emfit

end module at_experiments_mod
