!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_minimize_mod
!> @brief   perform energy minimization
!! @authors Takaharu Mori (TM), Chigusa Kobayashi (CK), Norio Takase (NT)
!!          Yoshinobu Akinaga (YA), Kiyoshi Yagi (KY), Yuji Sugita (YS)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_minimize_mod

  use at_output_mod
  use at_dynvars_mod
  use at_boundary_mod
  use at_pairlist_mod
  use at_energy_mod
  use at_output_str_mod
  use at_minimize_str_mod
  use at_dynvars_str_mod
  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use molecules_str_mod
  use fileio_rst_mod
  use timers_mod
  use fileio_control_mod
  use select_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use at_qmmm_mod

  implicit none
  private

  ! structures
  type, public :: s_min_info
    integer          :: method              = MinimizeMethodSD
    integer          :: nsteps              = 100
    integer          :: eneout_period       =  10
    integer          :: crdout_period       =   0
    integer          :: rstout_period       =   0
    integer          :: nbupdate_period     =  10
    integer          :: qmsave_period       =   1
    character(256)   :: fixatm_select_index = ''
    real(wp)         :: tol_rmsg            = -1.0_wp
    real(wp)         :: tol_maxg            = -1.0_wp
    logical          :: verbose             = .false.

    ! For SD
    real(wp)         :: force_scale_init    = 0.01_wp
    real(wp)         :: force_scale_max     = 0.1_wp

    ! For LBFGS
    integer          :: ncorrection         = 10
    logical          :: lbfgs_bnd           = .true.
    logical          :: lbfgs_bnd_qmonly    = .true.
    real(wp)         :: lbfgs_bnd_maxmove   = 0.1_wp

    ! For LBFGS - micro-iteration
    logical          :: macro               = .false.
    integer          :: nsteps_micro        = 100
    real(wp)         :: tol_rmsg_micro      = -1.0_wp
    real(wp)         :: tol_maxg_micro      = -1.0_wp
    character(256)   :: macro_select_index  = ''
    logical          :: init_lbfgs_micro    = .true.

  end type s_min_info

  ! subroutines
  public  :: show_ctrl_minimize
  public  :: read_ctrl_minimize
  public  :: setup_minimize
  private :: setup_atoms
  private :: print_atoms_info
  public  :: run_min
  public  :: steepest_descent
  public  :: minimize_lbfgs
  public  :: macro_iter
  public  :: micro_iter
  public  :: add_fixatm
  private :: delete_element

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_minimize
  !> @brief        show usage of MINIMIZE section
  !! @authors      NT, KY
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd", "rpath"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_minimize(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('min')

        write(MsgOut,'(A)') '[MINIMIZE]'
        write(MsgOut,'(A)') 'method            = SD'
        write(MsgOut,'(A)') 'nsteps            = 100      # number of minimization steps'
        write(MsgOut,'(A)') '# eneout_period   = 10       # energy output period'
        write(MsgOut,'(A)') '# crdout_period   = 0        # coordinates output period'
        write(MsgOut,'(A)') '# rstout_period   = 0        # restart output period'
        write(MsgOut,'(A)') '# nbupdate_period = 10       # nonbond update period'
        write(MsgOut,'(A)') '# verbose         = NO       # output verbosly'
        write(MsgOut,'(A)') '# force_scale_init = 0.01D+00  # initial scaling for SD'
        write(MsgOut,'(A)') '# force_scale_max  = 0.1D+00   # max scaling for SD'
        write(MsgOut,'(A)') ' '

      end select

    else

      select case (run_mode)

      case ('min')

        write(MsgOut,'(A)') '[MINIMIZE]'
        write(MsgOut,'(A)') 'method        = SD'
        write(MsgOut,'(A)') 'nsteps        = 100       # number of minimization steps'
        write(MsgOut,'(A)') ' '

      end select


    end if

    return

  end subroutine show_ctrl_minimize
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_minimize
  !> @brief        read MINIMIZE section in the control file
  !! @authors      TM, KY
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   min_info : MINIMIZE section in control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_minimize(handle, min_info)

    ! parameters
    character(*),            parameter     :: Section = 'Minimize'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_min_info),        intent(inout) :: min_info

    ! local
    logical :: found_error


    ! read parameters from control file
    !
    call begin_ctrlfile_section(handle, Section)

    call read_ctrlfile_type   (handle, Section, 'method',                   &
                               min_info%method, MinimizeMethodTypes)
    call read_ctrlfile_integer(handle, Section, 'nsteps',                   &
                               min_info%nsteps)
    call read_ctrlfile_integer(handle, Section, 'eneout_period',            &
                               min_info%eneout_period)
    call read_ctrlfile_integer(handle, Section, 'crdout_period',            &
                               min_info%crdout_period)
    call read_ctrlfile_integer(handle, Section, 'rstout_period',            &
                               min_info%rstout_period)
    call read_ctrlfile_integer(handle, Section, 'nbupdate_period',          &
                               min_info%nbupdate_period)
    call read_ctrlfile_integer(handle, Section, 'qmsave_period',            &
                               min_info%qmsave_period)
    call read_ctrlfile_string (handle, Section, 'fixatm_select_index',      &
                               min_info%fixatm_select_index)
    call read_ctrlfile_real   (handle, Section, 'tol_rmsg',                 &
                               min_info%tol_rmsg)
    call read_ctrlfile_real   (handle, Section, 'tol_maxg',                 &
                               min_info%tol_maxg)
    call read_ctrlfile_logical(handle, Section, 'verbose',                  &
                               min_info%verbose)

    call read_ctrlfile_real   (handle, Section, 'force_scale_max',          &
                               min_info%force_scale_max)
    call read_ctrlfile_real   (handle, Section, 'force_scale_init',         &
                               min_info%force_scale_init)

    call end_ctrlfile_section(handle)


    ! set default values for tol_rmsg/maxg
    !
    if (min_info%tol_rmsg < 0.0 .and. min_info%tol_maxg < 0.0) then
      min_info%tol_rmsg = 0.36_wp
      min_info%tol_maxg = min_info%tol_rmsg * 1.5_wp

    else if (min_info%tol_maxg < 0.0) then
      min_info%tol_maxg = min_info%tol_rmsg * 1.5_wp

    else if (min_info%tol_rmsg < 0.0) then
      min_info%tol_rmsg = min_info%tol_maxg * 2.0_wp / 3.0_wp

    end if

    if (min_info%tol_rmsg_micro < 0.0 .and. min_info%tol_maxg_micro < 0.0) then
      min_info%tol_rmsg_micro = min_info%tol_rmsg * 0.75_wp
      min_info%tol_maxg_micro = min_info%tol_maxg * 0.75_wp

    else if (min_info%tol_maxg_micro < 0.0) then
      min_info%tol_maxg_micro = min_info%tol_rmsg_micro * 1.5_wp

    else if (min_info%tol_rmsg_micro < 0.0) then
      min_info%tol_rmsg_micro = min_info%tol_maxg_micro * 2.0_wp / 3.0_wp

    end if


    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Parameters of MIN'

      write(MsgOut,'(A,A10,A,I10)')       &
            '  method                     = ', MinimizeMethodTypes(min_info%method), &
            '  nsteps                     = ', min_info%nsteps
      write(MsgOut,'(A,I10,A,I10)')       &
            '  eneout_period              = ', min_info%eneout_period, &
            '  crdout_period              = ', min_info%crdout_period
      write(MsgOut,'(A,I10,A,I10)')       &
            '  rstout_period              = ', min_info%rstout_period, &
            '  nbupdate_period            = ', min_info%nbupdate_period
      if (min_info%verbose) then
        write(MsgOut,'(A)')               &
            '  verbose                    =        yes'
      else
        write(MsgOut,'(A)')               &
            '  verbose                    =         no'
      end if

      if (min_info%method == MinimizeMethodSD) then
        write(MsgOut,'(A,F10.3,A,F10.3)') &
            '  force_scale_init           = ', min_info%force_scale_init, &
            '  force_scale_max            = ', min_info%force_scale_max
      end if

      write(MsgOut,'(A)') ' '
    end if


    ! error check
    !
    if (main_rank) then
      found_error = .false.

      if (min_info%eneout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%eneout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in eneout_period'
        write(MsgOut,'(A)') '  mod(nsteps, eneout_period) is not ZERO'
        found_error = .true.
      end if

      if (min_info%crdout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%crdout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in crdout_period'
        write(MsgOut,'(A)') '  mod(nsteps, crdout_period) is not ZERO'
        found_error = .true.
      end if

      if (min_info%rstout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%rstout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in rstout_period'
        write(MsgOut,'(A)') '  mod(nsteps, rstout_period) is not ZERO'
        found_error = .true.
      end if

      if (min_info%nbupdate_period > 0 .and.                      &
          mod(min_info%nsteps, min_info%nbupdate_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in nbupdate_period'
        write(MsgOut,'(A)') '  mod(nsteps, nbupdate_period) is not ZERO'
        found_error = .true.
      end if

      if (found_error) then
        call error_msg
      end if

    end if


    return

  end subroutine read_ctrl_minimize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_minimize
  !> @brief        setup minimize information
  !> @authors      TM, KY
  !! @param[in]    min_info : MINIMIZE section in control parameters
  !! @param[in]    out_info : OUTPUT   section in control parameters
  !! @param[in]    sel_info : SELECTOR section in control parameters
  !! @param[in]    molecule : molecule information
  !! @param[in]    qmmm     : qmmm information in enefunc
  !! @param[out]   minimize : minimize information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_minimize(min_info, out_info, sel_info, molecule, qmmm, minimize)

    ! formal arguments
    type(s_min_info),        intent(in)    :: min_info
    type(s_out_info),        intent(in)    :: out_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm),            intent(in)    :: qmmm
    type(s_minimize),        intent(inout) :: minimize

    ! local
    integer :: nwork, n3, iatom


    ! initialize variables in minimize
    !
    call init_minimize(minimize)


    ! setup minimize
    !
    minimize%method           = min_info%method
    minimize%nsteps           = min_info%nsteps
    minimize%eneout_period    = min_info%eneout_period
    minimize%crdout_period    = min_info%crdout_period
    minimize%rstout_period    = min_info%rstout_period
    minimize%nbupdate_period  = min_info%nbupdate_period
    minimize%qmsave_period    = min_info%qmsave_period
    minimize%tol_rmsg         = min_info%tol_rmsg
    minimize%tol_maxg         = min_info%tol_maxg
    minimize%verbose          = min_info%verbose

    if (out_info%dcdfile /= '' .and. minimize%crdout_period <= 0) &
      minimize%crdout_period = minimize%nsteps
    if (out_info%rstfile /= '' .and. minimize%rstout_period <= 0) &
      minimize%rstout_period = minimize%nsteps

    minimize%eneout           = .false.
    minimize%crdout           = .false.
    minimize%rstout           = .false.

    ! SD
    if (minimize%method == MinimizeMethodSD) then
      minimize%force_scale_init = min_info%force_scale_init
      minimize%force_scale_max  = min_info%force_scale_max
    end if

    ! setup atoms
    call setup_atoms(min_info, sel_info, molecule, qmmm, minimize)

    ! print_info
    call print_atoms_info(molecule, minimize)

    return

  end subroutine setup_minimize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_atoms
  !> @brief        define fixed atoms and opt atoms
  !! @authors      YA, KY
  !! @param[in]    min_info    : minimize input information
  !! @param[in]    sel_info    : selector input information
  !! @param[in]    molecule    : molecular information
  !! @param[in]    qmmm        : QM/MM information
  !! @param[out]   minimize    : minimize parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_atoms(min_info, sel_info, molecule, qmmm, minimize)

    ! formal arguments
    type(s_min_info),        intent(in) :: min_info
    type(s_sel_info),        intent(in) :: sel_info
    type(s_molecule),        intent(in) :: molecule
    type(s_qmmm),            intent(in) :: qmmm
    type(s_minimize),     intent(inout) :: minimize

    ! local variables
    integer                :: igroup, ngroup, natom, i, j, offset, temp, iatom
    integer, allocatable   :: group_list(:)
    type(s_selatoms), allocatable :: selatoms(:)

    integer, allocatable   :: input_id(:)

    integer :: ntmp, num_macro, num_micro
    integer, allocatable :: itmp_macro(:), itmp_micro(:), itmp_qmmmbond_list(:)

    integer :: ifix, iopt, iend

    ! Set up fixed atoms
    !
    ngroup = split_num(trim(min_info%fixatm_select_index))
    if (ngroup /= 0) then

      ! Number of fixed atoms
      !
      allocate(group_list(ngroup))
      call split(ngroup, ngroup, min_info%fixatm_select_index, group_list)

      allocate(selatoms(ngroup))

      natom = 0
      do i = 1, ngroup
        igroup = group_list(i)
        call select_atom(molecule, sel_info%groups(igroup), selatoms(i))
        natom = natom + size(selatoms(i)%idx)
      end do

      minimize%num_fixatoms = natom

      ! List of fixed atoms
      !
      allocate(minimize%fixedatom_id(minimize%num_fixatoms))

      offset = 0
      do i = 1, ngroup
        igroup = group_list(i)
        natom = size(selatoms(i)%idx)
        minimize%fixedatom_id(offset+1:offset+natom) = selatoms(i)%idx(1:natom)
        offset = offset + natom
      end do

      deallocate(selatoms)
      deallocate(group_list)

      ! Sort fixed atom indices in ascending order
      !
      do i = minimize%num_fixatoms, 2, -1
        do j = 1, i - 1
          if (minimize%fixedatom_id(j) > minimize%fixedatom_id(j+1)) then
            temp = minimize%fixedatom_id(j)
            minimize%fixedatom_id(j)   = minimize%fixedatom_id(j+1)
            minimize%fixedatom_id(j+1) = temp
          end if
        end do
      end do

      ! Set up opt atoms
      !
      minimize%num_optatoms = molecule%num_atoms - minimize%num_fixatoms
      allocate(minimize%optatom_id(minimize%num_optatoms))

      ifix = 1
      iopt = 1
      iend = -1
      do i = 1, molecule%num_atoms
        if (i == minimize%fixedatom_id(ifix)) then
          ifix = ifix + 1
          if (ifix > minimize%num_fixatoms) then 
            iend = i
            exit
          end if

        else
          minimize%optatom_id(iopt) = i
          iopt = iopt + 1
        end if
      end do

      if (iend > 0) then
        do i = iend + 1, molecule%num_atoms
          minimize%optatom_id(iopt) = i
          iopt = iopt + 1
        end do
      end if

    else

      ! No fixed atoms
      !
      minimize%num_fixatoms = 0

      ! Set up opt atoms
      !
      minimize%num_optatoms = molecule%num_atoms
      allocate(minimize%optatom_id(minimize%num_optatoms))
      do i = 1, molecule%num_atoms
        minimize%optatom_id(i) = i
      end do

    end if

    ! Set up opt atoms for macro/micro-iterations
    ! 
    if (minimize%macro) then
 
      ntmp = minimize%num_optatoms
      allocate(itmp_macro(ntmp), itmp_micro(ntmp))
      itmp_micro = minimize%optatom_id
      num_micro  = minimize%num_optatoms
      itmp_macro = 0
      num_macro  = 0

      ! delete QM atoms
      call delete_element(qmmm%qm_natoms, qmmm%qmatom_id, &
                          num_micro, itmp_micro,          &
                          ntmp, itmp_macro(num_macro+1:) )
      num_macro = num_macro + ntmp

      ! delete link MM atoms
      if (qmmm%num_qmmmbonds > 0) then
        allocate(itmp_qmmmbond_list(qmmm%num_qmmmbonds))

        itmp_qmmmbond_list = qmmm%qmmmbond_list(2,:)

        ! Sort atom indices in ascending order
        !
        do i = qmmm%num_qmmmbonds, 2, -1
          do j = 1, i - 1
            if (itmp_qmmmbond_list(j) > itmp_qmmmbond_list(j+1)) then
              temp = itmp_qmmmbond_list(j)
              itmp_qmmmbond_list(j)   = itmp_qmmmbond_list(j+1)
              itmp_qmmmbond_list(j+1) = temp
            end if
          end do
        end do

        call delete_element(qmmm%num_qmmmbonds, itmp_qmmmbond_list, &
                            num_micro, itmp_micro,          &
                            ntmp, itmp_macro(num_macro+1:) )
        num_macro = num_macro + ntmp

        deallocate(itmp_qmmmbond_list)
      endif

      ! Get input atom ID
      ! 
      ngroup = split_num(trim(min_info%macro_select_index))
      if (ngroup /= 0) then
        allocate(group_list(ngroup))
        call split(ngroup, ngroup, min_info%macro_select_index, group_list)
 
        allocate(selatoms(ngroup))
 
        natom = 0
        do i = 1, ngroup
          igroup = group_list(i)
          call select_atom(molecule, sel_info%groups(igroup), selatoms(i))
          natom = natom + size(selatoms(i)%idx)
        end do
 
        allocate(input_id(natom))
 
        offset = 0
        do i = 1, ngroup
          igroup = group_list(i)
          natom = size(selatoms(i)%idx)
          input_id(offset+1:offset+natom) = selatoms(i)%idx(1:natom)
          offset = offset + natom
        end do
 
        deallocate(selatoms)
        deallocate(group_list)
 
        ! Sort input atom indices in ascending order
        !
        do i = natom, 2, -1
          do j = 1, i - 1
            if (input_id(j) > input_id(j+1)) then
              temp = input_id(j)
              input_id(j)   = input_id(j+1)
              input_id(j+1) = temp
            end if
          end do
        end do

        ! delete input atoms
        call delete_element(natom, input_id, num_micro, itmp_micro,   &
                            ntmp, itmp_macro(num_macro+1:))
        num_macro = num_macro + ntmp

      end if

      minimize%num_optatoms_micro = num_micro
      allocate(minimize%optatom_micro_id(minimize%num_optatoms_micro))
      do i = 1, minimize%num_optatoms_micro
        minimize%optatom_micro_id(i) = itmp_micro(i)
      end do

      minimize%num_optatoms_macro = num_macro
      allocate(minimize%optatom_macro_id(minimize%num_optatoms_macro))
      do i = 1, minimize%num_optatoms_macro
        minimize%optatom_macro_id(i) = itmp_macro(i)
      end do

      ! Sort atom indices in ascending order
      !
      do i = num_macro, 2, -1
        do j = 1, i - 1
          if (minimize%optatom_macro_id(j) > minimize%optatom_macro_id(j+1)) then
            temp = minimize%optatom_macro_id(j)
            minimize%optatom_macro_id(j)   = minimize%optatom_macro_id(j+1)
            minimize%optatom_macro_id(j+1) = temp
          end if
        end do
      end do

      deallocate(itmp_micro, itmp_macro)

    end if

    return

  end subroutine setup_atoms

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    print_atoms_info
  !> @brief        print information of atoms
  !! @authors      KY
  !! @param[in]    min_info    : minimize input information
  !! @param[in]    sel_info    : selector input information
  !! @param[in]    molecule    : molecular information
  !! @param[in]    qmmm        : QM/MM information
  !! @param[out]   minimize    : minimize parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine print_atoms_info(molecule, minimize)

    ! formal arguments
    type(s_molecule),        intent(in) :: molecule
    type(s_minimize),     intent(inout) :: minimize

    ! local variables
    integer :: i, iatom
    integer, parameter :: max_atm_print = 100

    ! Punch out opt atoms info.
    !
    if (main_rank) then
      write(MsgOut,'(a)') "Setup_MinAtoms> Atoms info in minimize"
      if (minimize%num_fixatoms < max_atm_print) then
        do i = 1, minimize%num_fixatoms
          iatom   = minimize%fixedatom_id(i)
          write(MsgOut,'(i6,x,a4,x,i6,x,a4,x,a6,x,a6)') &
            iatom, &
            molecule%segment_name(iatom), &
            molecule%residue_no(iatom),   &
            molecule%residue_name(iatom), &
            molecule%atom_name(iatom),    &
            molecule%atom_cls_name(iatom)
        end do
      end if
      write(MsgOut,'(a,i0)') "  number of fixed atoms     = ", minimize%num_fixatoms
      if (minimize%num_optatoms < max_atm_print) then
        do i = 1, minimize%num_optatoms
          iatom   = minimize%optatom_id(i)
          write(MsgOut,'(i6,x,a4,x,i6,x,a4,x,a6,x,a6)') &
            iatom, &
            molecule%segment_name(iatom), &
            molecule%residue_no(iatom),   &
            molecule%residue_name(iatom), &
            molecule%atom_name(iatom),    &
            molecule%atom_cls_name(iatom)
        end do
      end if
      write(MsgOut,'(a,i0)') "  number of minimized atoms = ", minimize%num_optatoms
      write(MsgOut, '(a)') ' '

      if (minimize%macro) then
        if (minimize%num_optatoms_macro > 0 .and. &
            minimize%num_optatoms_macro < max_atm_print) then
          write(MsgOut,'(a)') "  Atoms optimized during macro-iteration"
          do i = 1, minimize%num_optatoms_macro
            iatom   = minimize%optatom_macro_id(i)
            write(MsgOut,'(i10,x,a4,x,i6,x,a4,x,a6,x,a6)') &
              iatom, &
              molecule%segment_name(iatom), &
              molecule%residue_no(iatom),   &
              molecule%residue_name(iatom), &
              molecule%atom_name(iatom),    &
              molecule%atom_cls_name(iatom)
          end do
        end if
        write(MsgOut,'(a,i0)') "  number of atoms in macro-iteration = ", &
              minimize%num_optatoms_macro
        write(MsgOut,'(a,i0)') "  number of atoms in micro-iteration = ", &
              minimize%num_optatoms_micro
        write(MsgOut, '(a)') ' '
      end if
    end if

  end subroutine print_atoms_info

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_min
  !> @brief        perform energy minimization
  !! @authors      YS
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions
  !! @param[inout] dynvars     : dynamic variables
  !! @param[inout] minimize    : minimize information
  !! @param[inout] pairlist    : non-bond pair list
  !! @param[inout] boundary    : boundary condition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_min(output, molecule, enefunc, dynvars, minimize, &
                     pairlist, boundary)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_minimize),         intent(inout) :: minimize
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary

    ! Open output files
    !
    call open_output(output)

    ! MIN main loop
    !
    select case (minimize%method)

    case (MinimizeMethodSD)

      call steepest_descent (output, molecule, enefunc, dynvars, minimize, &
                             pairlist, boundary)

    case (MinimizeMethodLBFGS)

      if (.not. minimize%macro) then
        call minimize_lbfgs (output, molecule, enefunc, dynvars, minimize, &
                             pairlist, boundary)
      else
        call macro_iter (output, molecule, enefunc, dynvars, minimize, &
                         pairlist, boundary)
      end if

      call dealloc_minimize(minimize, MinimizeLBFGS)

    end select


    return

  end subroutine run_min

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    steepest_descent
  !> @brief        steepest descent integrator
  !> @authors      TM, CK, KY
  !! @param[inout] output   : output information
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] dynvars  : dynamic variables information
  !! @param[inout] minimize : minimize information
  !! @param[inout] pairlist : non-bond pair list information
  !! @param[inout] boundary : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine steepest_descent(output, molecule, enefunc, dynvars, minimize, &
                              pairlist, boundary)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_minimize), target, intent(inout) :: minimize
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary

    ! local variables
    real(wp)                  :: energy_ref, delta_energy
    real(wp)                  :: delta_r, delta_rini, delta_rmax
    real(wp)                  :: rmsg, maxg, absg
    integer                   :: i, j, jid, k, natom, natom_opt, no3, nsteps
    integer, pointer          :: optatom_id(:)

    real(wp),         pointer :: coord(:,:), coord_ref(:,:), force(:,:)

    character(256) :: folder,basename
    character      :: num*5
    logical        :: savefile


    ! use pointers
    !
    coord      => dynvars%coord
    coord_ref  => dynvars%coord_ref
    force      => dynvars%force

    nsteps     =  minimize%nsteps
    natom      =  molecule%num_atoms
    natom_opt  = minimize%num_optatoms
    optatom_id => minimize%optatom_id
    no3       = natom_opt*3

    delta_rini = minimize%force_scale_init
    delta_rmax = minimize%force_scale_max
    delta_r    = delta_rini

    ! qmmm settings
    if(enefunc%qmmm%do_qmmm) then
      folder   = 'qmmm_mini'
      basename = 'min00000'
      savefile = .true.
      call set_runtime_qmmm(enefunc%qmmm, folder, basename, savefile)
      enefunc%qmmm%qm_classical = .false.
    end if

    ! Compute energy of the initial structure
    !
    call compute_energy(molecule, enefunc, pairlist, boundary, .true., &
                        enefunc%nonb_limiter,  &
                        dynvars%coord,         &
                        dynvars%energy,        &
                        dynvars%temporary,     &
                        dynvars%force,         &
                        dynvars%virial,        &
                        dynvars%virial_extern)

    rmsg = 0.0_wp
    maxg = 0.0_wp
    do j = 1, natom_opt
      jid  = optatom_id(j)
      do k = 1, 3
        rmsg = rmsg + force(k,jid)**2
        absg = abs(force(k,jid))
        if(absg > maxg) maxg = absg
      end do
    end do
    rmsg = sqrt(rmsg/real(no3,wp))
    dynvars%rms_gradient = rmsg
    dynvars%max_gradient = maxg

    do j = 1, natom_opt
      jid  = optatom_id(j)
      coord(1:3,jid) = coord(1:3,jid) + delta_r*force(1:3,jid)/rmsg
    end do

    dynvars%step = 0
    dynvars%total_pene = dynvars%energy%total

    call output_dynvars(output, enefunc, dynvars)


    ! Main loop of the Steepest descent method
    !
    do i = 1, nsteps

      dynvars%step = i

      ! save old energy and coordinates
      !
      do j = 1, natom_opt
        jid  = optatom_id(j)
        coord_ref(1:3,jid) = coord(1:3,jid)
      end do
      energy_ref = dynvars%energy%total

      ! qmmm settings
      if(enefunc%qmmm%do_qmmm) then
        folder   = 'qmmm_mini'
        write(num,'(i5.5)') i
        basename = 'min'//num

        if(mod(i,minimize%qmsave_period) == 0) then
          savefile = .true.
        else
          savefile = .false.
        endif
        call set_runtime_qmmm(enefunc%qmmm, folder, basename, savefile)
        enefunc%qmmm%qm_classical = .false.
      end if

      ! Compute energy and forces
      !
      call compute_energy(molecule, enefunc, pairlist, boundary,  &
                          .true.,                &
                          enefunc%nonb_limiter,  &
                          dynvars%coord,         &
                          dynvars%energy,        &
                          dynvars%temporary,     &
                          dynvars%force,         &
                          dynvars%virial,        &
                          dynvars%virial_extern)

      delta_energy = dynvars%energy%total - energy_ref
      if (delta_energy > 0) then
        delta_r = 0.5_wp*delta_r
      else
        delta_r = min(delta_rmax, 1.2_wp*delta_r)
      end if

      ! check RMSG and MAXG convergence
      !
      rmsg = 0.0_wp
      maxg = 0.0_wp
      do j = 1, natom_opt
        jid  = optatom_id(j)
        do k = 1, 3
          rmsg = rmsg + force(k,jid)**2
          absg = abs(force(k,jid))
          if(absg > maxg) maxg = absg
        end do
      end do
      rmsg = sqrt(rmsg/real(no3,wp))
      dynvars%rms_gradient = rmsg
      dynvars%max_gradient = maxg

      if (rmsg <= minimize%tol_rmsg .and. maxg <= minimize%tol_maxg) then
        minimize%eneout = .true.
        minimize%crdout = .true.
        minimize%rstout = .true.
        call output_min(output, molecule, enefunc, minimize, boundary, &
                        delta_r, dynvars)
        if (enefunc%qmmm%do_qmmm .and. enefunc%qmmm%save_qminfo) &
          call write_qminfo(enefunc%qmmm, dynvars%energy%qm_ene)
        if (main_rank) &
          write(MsgOut, '(" Steepest Descent converged: RMSG sufficiently small")')
        exit
      end if

      ! update structure
      !
      do j = 1, natom_opt
        jid  = optatom_id(j)
        coord(1:3,jid) = coord(1:3,jid) + delta_r*force(1:3,jid)/rmsg
      end do

      ! update nonbond pairlist
      !
      if (minimize%nbupdate_period > 0) then
        if (mod(i,minimize%nbupdate_period) == 0 .and. real_calc) then

          call update_pairlist(enefunc, boundary, coord, pairlist)

        end if
      end if


      ! output energy and dynamical variables
      !
     call output_min(output, molecule, enefunc, minimize, boundary, &
                     delta_r, dynvars)


    end do


    ! Close output files
    !
    call close_output(output)


    return

  end subroutine steepest_descent

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine : minimize_lbfgs
  !> @brief  minimzation with lbfgs
  !> @author CK, KY, YA
  !! @param[inout] output:     information about output [str]
  !! @param[inout] molecule:   information about molecules [str]
  !! @param[inout] enefunc:    potential energy function [str]
  !! @param[inout] dynvars:    dynamics varibales [str]
  !! @param[inout] minimize:   information about minimize [str]
  !! @param[inout] pairlist:   pairlist [str]
  !! @param[inout] boundary:   boundary condition [str]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine minimize_lbfgs (output, molecule, enefunc, dynvars, minimize, &
                             pairlist, boundary)

    ! formal arguments
    type(s_output),            intent(inout) :: output
    type(s_molecule), target,  intent(inout) :: molecule
    type(s_enefunc),           intent(inout) :: enefunc
    type(s_dynvars),  target,  intent(inout) :: dynvars
    type(s_minimize), target,  intent(inout) :: minimize
    type(s_pairlist),          intent(inout) :: pairlist
    type(s_boundary),          intent(inout) :: boundary

    ! local variables
    integer                :: i, j, k, ii
    real(wp)               :: energy1
    real(wp)               :: rmsg, maxg, absg
    integer                :: maxg_id

    integer                :: natom
    integer                :: natom_opt, no3
    integer, pointer       :: optatom_id(:)
    integer                :: nsteps, ncount

    integer                :: ncorr, nwork
    character(len=60)      :: task, csave
    logical                :: lsave(4)
    integer                :: isave(44)
    real(wp)               :: dsave(29)
    real(wp), allocatable  :: lower(:), upper(:)
    real(wp), allocatable  :: vec(:), gradient(:), work_lbfgs(:)
    integer , allocatable  :: list_bound(:), iwork_lbfgs(:)
    integer                :: iprint = -1
    real(wp)               :: factr, pgtol
    logical                :: bnd, bnd_qmonly
    real(wp)               :: maxmove

    real(wp)   , pointer   :: coord(:,:), coord_ref(:,:), force(:,:)

    character(256) :: folder,basename
    character      :: num*5
    logical        :: savefile

    integer                :: outunit

    real(wp)               :: delta_r


    return

  end subroutine minimize_lbfgs

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    macro_iter
  !> @brief        macro-iteration for QM atoms with QM/MM energy
  !> @authors      YA, KY
  !! @param[inout] output   : information about output [str]
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] dynvars  : dynamic variables information
  !! @param[inout] minimize : minimize information
  !! @param[inout] pairlist : non-bond pair list information
  !! @param[inout] boundary : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine macro_iter (output, molecule, enefunc, dynvars, minimize, &
                         pairlist, boundary)

    ! formal arguments
    type(s_output),            intent(inout) :: output
    type(s_molecule), target,  intent(inout) :: molecule
    type(s_enefunc),           intent(inout) :: enefunc
    type(s_dynvars),  target,  intent(inout) :: dynvars
    type(s_minimize), target,  intent(inout) :: minimize
    type(s_pairlist),          intent(inout) :: pairlist
    type(s_boundary),          intent(inout) :: boundary

    ! local variables
    integer                :: i, j, k, ii
    real(wp)               :: energy1 = 0.0_wp
    real(wp)               :: rmsg, maxg, absg
    integer                :: maxg_id

    integer                :: natom
    integer                :: natom_opt, natom_macro, natom_micro, no3
    integer, pointer       :: optatom_id(:)
    integer, pointer       :: optatom_macro_id(:)
    integer, pointer       :: optatom_micro_id(:)
    integer                :: nsteps, ncount

    integer                :: ncorr, nwork
    character(len=60)      :: task, csave
    logical                :: lsave(4)
    integer                :: isave(44)
    real(wp)               :: dsave(29)
    real(wp), allocatable  :: lower(:), upper(:)
    real(wp), allocatable  :: vec(:), gradient(:), work_lbfgs(:)
    integer , allocatable  :: list_bound(:), iwork_lbfgs(:)
    integer                :: iprint = -1
    real(wp)               :: factr, pgtol
    logical                :: bnd, bnd_qmonly
    real(wp)               :: maxmove

    real(wp)    , pointer  :: coord(:,:), coord_ref(:,:), force(:,:)

    character(256) :: folder,basename
    character      :: num*5
    logical        :: savefile

    real(wp)              :: energy0corr
    real(wp), allocatable :: coord0(:,:), force0corr(:,:)
    
    real(wp)              :: delta_r

    return

  end subroutine macro_iter

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    micro_iter
  !> @brief        micro-iteration for MM atoms with MM energy
  !> @authors      YA, KY
  !! @param[inout] molecule  : molecule information
  !! @param[inout] enefunc   : potential energy functions information
  !! @param[inout] dynvars   : dynamic variables information
  !! @param[inout] minimize  : minimize information
  !! @param[inout] pairlist  : non-bond pair list information
  !! @param[inout] boundary  : boundary conditions information
  !! @param[in]  energy0corr : correction of the energy
  !! @param[in]  force0corr  : correction of the force
  !! @param[in]  iter        : counter of the macro iteration
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine micro_iter(output, molecule, enefunc, dynvars, minimize, &
                        pairlist, boundary, coord0, energy0corr, force0corr, iter)

    ! formal arguments
    type(s_output),            intent(inout) :: output
    type(s_molecule), target,  intent(inout) :: molecule
    type(s_enefunc),           intent(inout) :: enefunc
    type(s_dynvars),  target,  intent(inout) :: dynvars
    type(s_minimize), target,  intent(inout) :: minimize
    type(s_pairlist),          intent(inout) :: pairlist
    type(s_boundary),          intent(inout) :: boundary
    real(wp),                  intent(in)    :: coord0(3,*)
    real(wp),                  intent(in)    :: energy0corr
    real(wp),                  intent(in)    :: force0corr(3,*)
    integer,                   intent(in)    :: iter

    ! local variables
    integer                :: i, j, k
    real(wp)               :: rmsg, maxg, absg
    logical                :: print_log

    integer                :: natom
    integer                :: natom_opt, no3
    integer, pointer       :: optatom_id(:)
    integer                :: nsteps, ncount

    integer                :: ncorr
    character(len=60)      :: task, csave
    logical     , pointer  :: lsave(:)
    integer     , pointer  :: isave(:)
    real(wp)    , pointer  :: dsave(:)
    real(wp)    , pointer  :: lower(:), upper(:)
    real(wp)    , pointer  :: vec(:), gradient(:), work_lbfgs(:)
    integer     , pointer  :: list_bound(:), iwork_lbfgs(:)
    integer                :: iprint = -1
    real(wp)               :: factr, pgtol
    logical                :: bnd, bnd_qmonly
    real(wp)               :: maxmove

    real(wp)    , pointer  :: coord(:,:), coord_ref(:,:), force(:,:)

    integer                :: indx
    integer                :: outunit
    real(wp)               :: energy1, gdx

    real(wp)               :: delta_r

    return

  end subroutine micro_iter

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    add_fixatm
  !> @brief        add atoms to be fixed in MINIMIZATION
  !> @authors      KY
  !! @param[in]    natoms    : number of atoms
  !! @param[in]    atom_id   : list of atoms
  !! @param[in]    molecule  : molecule information
  !! @param[inout] minimize  : minimize information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine add_fixatm(natoms, atom_id, molecule, minimize)

    ! formal arguments
    integer         , intent(in)    :: natoms
    integer         , intent(in)    :: atom_id(natoms)
    type(s_molecule), intent(in)    :: molecule
    type(s_minimize), intent(inout) :: minimize

    ! local variables
    integer              :: nd, nf0, i, j, k
    integer, allocatable :: del_id(:), tmp(:)

    nd = minimize%num_optatoms
    allocate(del_id(nd))
    call delete_element(natoms, atom_id, minimize%num_optatoms, &
                        minimize%optatom_id, nd, del_id)

    if (nd == 0) return

    if(main_rank) then
      write(MsgOut,'(a)') "Setup_MinAtoms> Atoms info is changed in minimize"
      write(MsgOut,'(a,i0)') "  number of atoms deleted from opt = ", nd
      write(MsgOut,*)
    end if

    ! reallocate fixedatom_id
    if (minimize%num_fixatoms > 0) then
      allocate(tmp(minimize%num_fixatoms))
      tmp = minimize%fixedatom_id
      nf0 = minimize%num_fixatoms

      deallocate(minimize%fixedatom_id)
      minimize%num_fixatoms = minimize%num_fixatoms + nd
      allocate(minimize%fixedatom_id(minimize%num_fixatoms))

      j = 1
      k = 1
      do i = 1, minimize%num_fixatoms
        if(tmp(j) < del_id(k)) then
          minimize%fixedatom_id(i) = tmp(j)
          j = j + 1
          if (j > nf0) then
            minimize%fixedatom_id(i+1:minimize%num_fixatoms) = &
              del_id(k:nd)
            exit
          end if
        else
          minimize%fixedatom_id(i) = del_id(k)
          k = k + 1
          if (k > nd) then
            minimize%fixedatom_id(i+1:minimize%num_fixatoms) = &
              tmp(j:nf0)
            exit
          end if
        end if
      end do

      deallocate(tmp)

    else
      minimize%num_fixatoms = nd
      allocate(minimize%fixedatom_id(minimize%num_fixatoms))
      minimize%fixedatom_id = del_id

    end if

    if (minimize%macro) then
      call delete_element(natoms, atom_id, minimize%num_optatoms_macro, &
                          minimize%optatom_macro_id, nd, del_id)
      call delete_element(natoms, atom_id, minimize%num_optatoms_micro, &
                          minimize%optatom_micro_id, nd, del_id)
    end if

    call print_atoms_info(molecule, minimize)

    deallocate(del_id)

  end subroutine add_fixatm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    delete_element
  !> @brief        deletes an element of vector b if it is an element of
  !>               vector a. Note that a and b are assumed to be in ascending 
  !>               order, i.e., a(1) < a(2) < ... and b(1) < b(2) < ...
  !> @authors      KY
  !! @param[in]    na     : number of elements of a
  !! @param[in]    a(na)  : vector a
  !! @param[inout] nb     : number of elements of b
  !! @param[inout] b(nb)  : vector b
  !! @param[inout] nd     : number of deleted elements
  !! @param[inout] d(nd)  : vector of deleted elements
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine delete_element(na, a, nb, b, nd, d)

    ! formal arguments
    integer, intent(in) :: na
    integer, intent(in) :: a(na)
    integer, intent(inout) :: nb
    integer, intent(inout) :: b(*)
    integer, optional, intent(inout) :: nd
    integer, optional, intent(inout) :: d(*)

    ! local variables
    integer              :: i, j, k, nb0
    integer, allocatable :: tmp(:)
    logical, allocatable :: del(:)

    allocate(del(nb))
    nd = 0
    do i = 1, nb
      del(i) = .false.
      do j = 1, na
        if (a(j) == b(i)) then
          nd = nd + 1
          del(i) = .true.
          exit
        else if (a(j) > b(i)) then
          exit
        end if
      end do
    end do

    if(nd == 0) then
      deallocate(del)
      return
    end if

    allocate(tmp(nb))
    tmp = b(1:nb)
    nb0 = nb
    nb = nb - nd

    j = 1
    k = 1
    do i = 1, nb0
      if(.not. del(i)) then
        b(j) = tmp(i)
        j = j + 1
      else
        d(k) = tmp(i)
        k = k + 1
      end if
    end do

    deallocate(del,tmp)

  end subroutine delete_element

end module at_minimize_mod
