!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_qmmm_mod
!> @brief   perform QM/MM calculation
!! @authors Yoshinobu Akinaga (YA), Kiyoshi Yagi (KY), Kenta Yamada (KYMD)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_qmmm_mod

  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use at_energy_str_mod
  use at_boundary_str_mod
  use molecules_str_mod
  use select_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use fileio_mod
  use fileio_control_mod
  use fileio_psf_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! atomic symbols
  character(3)            :: qm_atomic_symbol(118) = (/ &
      'H  ', 'He ', 'Li ', 'Be ', 'B  ', 'C  ', 'N  ', 'O  ', 'F  ', 'Ne ', &
      'Na ', 'Mg ', 'Al ', 'Si ', 'P  ', 'S  ', 'Cl ', 'Ar ', 'K  ', 'Ca ', &
      'Sc ', 'Ti ', 'V  ', 'Cr ', 'Mn ', 'Fe ', 'Co ', 'Ni ', 'Cu ', 'Zn ', &
      'Ga ', 'Ge ', 'As ', 'Se ', 'Br ', 'Kr ', 'Rr ', 'Sr ', 'Y  ', 'Zr ', &
      'Nb ', 'Mo ', 'Tc ', 'Ru ', 'Rh ', 'Pd ', 'Ag ', 'Cd ', 'In ', 'Sn ', &
      'Sb ', 'Te ', 'I  ', 'Xe ', 'Cs ', 'Ba ', 'La ', 'Ce ', 'Pr ', 'Nd ', &
      'Pm ', 'Sm ', 'Eu ', 'Gd ', 'Tb ', 'Dy ', 'Ho ', 'Er ', 'Tm ', 'Yb ', &
      'Lu ', 'Hf ', 'Ta ', 'W  ', 'Re ', 'Os ', 'Ir ', 'Pt ', 'Au ', 'Hg ', &
      'Tl ', 'Pb ', 'Bi ', 'Po ', 'At ', 'Rn ', 'Fr ', 'Ra ', 'Ac ', 'Th ', &
      'Pa ', 'U  ', 'Np ', 'Pu ', 'Am ', 'Cm ', 'Bk ', 'Cf ', 'Es ', 'Fm ', &
      'Md ', 'No ', 'Lr ', 'Rf ', 'Db ', 'Sg ', 'Bh ', 'Hs ', 'Mt ', 'Ds ', &
      'Rg ', 'Cn ', 'Uut', 'Fl ', 'Uup', 'Lv ', 'Uus', 'Uuo' /)

  ! atomic mass (Taken from Standard Atomic Weights 2013)
  real(wp)               :: qm_atomic_mass(118) = (/ &
      1.0080_wp,   4.0026_wp,   6.9675_wp,   9.0122_wp,  10.8135_wp, &
     12.0106_wp,  14.0069_wp,  15.9994_wp,  18.9984_wp,  20.1797_wp, &
     22.9898_wp,  24.3050_wp,  26.9815_wp,  28.0850_wp,  30.9738_wp, &
     32.0675_wp,  35.4515_wp,  39.948_wp,   39.0983_wp,  40.078_wp,  &
     44.9559_wp,  47.867_wp,   50.9415_wp,  51.9961_wp,  54.9380_wp, &
     55.845_wp,   58.9332_wp,  58.6934_wp,  63.546_wp,   65.38_wp,   &
     69.723_wp,   72.63_wp,    74.9216_wp,  78.96_wp,    79.904_wp,  &
     83.798_wp,   85.4678_wp,  87.62_wp,    88.9059_wp,  91.224_wp,  &
     92.9064_wp,  95.96_wp,     0.0000_wp, 101.07_wp,   102.9055_wp, &
    106.42_wp,   107.8682_wp, 112.411_wp,  114.818_wp,  118.710_wp,  &
    121.760_wp,  127.60_wp,   126.9044_wp, 131.293_wp,  132.9055_wp, &
    137.327_wp,  138.9055_wp, 140.116_wp,  140.9077_wp, 144.242_wp,  &
      0.0000_wp, 150.36_wp,   151.964_wp,  157.25_wp,   158.9254_wp, &
    162.500_wp,  164.9303_wp, 167.259_wp,  168.9342_wp, 173.054_wp,  &
    174.9668_wp, 178.49_wp,   180.9479_wp, 183.84_wp,   186.207_wp,  &
    190.23_wp,   192.217_wp,  195.084_wp,  196.9666_wp, 200.59_wp,   &
    204.384_wp,  207.2_wp,    208.9804_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, 232.0381_wp, &
    231.0359_wp, 238.0289_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp,   0.0000_wp, &
      0.0000_wp,   0.0000_wp,   0.0000_wp /)

  real(wp), parameter     :: deuterium_mass = 2.0141_wp
  !
  real(wp), parameter     :: mass_error = 0.05_wp

  ! structures
  type, public :: s_qmmm_info
    logical               :: do_qmmm = .false.
    integer               :: qmtyp = 0
    character(256)        :: qmcnt = ''
    character(256)        :: qmexe = ''
    character(256)        :: qmatm_select_index  = ''
    integer               :: exclude_charge      = ExcludeChargeGroup
    integer               :: qmmaxtrial          = 1
    logical               :: qm_debug            = .false.
    logical               :: qminfo              = .false.
    real(wp)              :: mm_cutoffdist       = -1.0_wp
    logical               :: mm_cutoff_byresidue = .true.

    character(256)        :: qm_exemode = 'system'
    integer               :: qm_nprocs  = 1
  end type s_qmmm_info

  ! subroutines
  public  :: show_ctrl_qmmm
  public  :: read_ctrl_qmmm
  public  :: setup_qmmm
  public  :: qm_atomic_number
  public  :: qm_generate_input
  public  :: qm_read_output
  public  :: write_qminfo
  public  :: qm_monitor
  private :: setup_mmlist
  public  :: update_mmlist
  private :: qm_check_input_qchem
  private :: qm_check_input_gaussian
  private :: qm_check_input_gaussian_frwf
  private :: qm_check_input_molpro
  private :: qm_generate_input_qchem
  private :: qm_generate_input_gaussian
  private :: qm_generate_input_gaussian_frwf
  private :: qm_generate_input_molpro
  private :: qm_read_output_qchem
  private :: qm_read_output_gaussian
  private :: qm_read_output_gaussian_frwf
  private :: qm_read_output_molpro
  private :: count_qmmmbonds
  private :: generate_linkatoms
  private :: check_linkmm
  private :: rearrange_term_list
  private :: rearrange_impr_list

contains


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_qmmm
  !> @brief        show QMMM section usage
  !! @authors      YA
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "remd", "min", "rpath"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_qmmm(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    return

  end subroutine show_ctrl_qmmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_qmmm
  !> @brief        read QMMM section in the control file
  !! @authors      YA, KY
  !! @param[in]    handle   : unit number of control fileha
  !! @param[in]    min_info : QMMM section in control parameters 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_qmmm(handle, ctrlfile, qmmm_info)

    ! parameters
    character(*),            parameter     :: Section = 'QMMM'

    ! formal arguments
    integer,                 intent(in)    :: handle
    character(*),            intent(in)    :: ctrlfile
    type(s_qmmm_info),       intent(inout) :: qmmm_info

    ! local variables
    character(256)           :: line
    character(6)             :: word
    logical                  :: found
    integer                  :: unit_no, nsta, nend


    return

  end subroutine read_ctrl_qmmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_qmmm
  !> @brief        setup QM/MM information
  !> @authors      YA
  !! @param[in]    qmmm_info : QMMM section control parameters information
  !! @param[in]    sel_info  : SELECTION section control parameters information
  !! @param[in]    boundary  : boundary information
  !! @param[in]    psf       : PSF information
  !! @param[in]    molecule  : molecular information
  !! @param[out]   qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_qmmm(qmmm_info, sel_info, boundary, psf, molecule, qmmm)

    ! formal arguments
    type(s_qmmm_info),       intent(in)    :: qmmm_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_psf),             intent(in)    :: psf
    type(s_boundary),        intent(in)    :: boundary
    type(s_molecule),        intent(inout) :: molecule
    type(s_qmmm),            intent(inout) :: qmmm

    ! local variables
    character(4)                  :: mm_name
    integer                       :: atomic_no
    integer                       :: i, j, offset
    integer                       :: ngroup, igroup, natom, iatom
    integer                       :: istart, iend, mm_start, mm_end, mm_block
    integer                       :: num_partner, list_partner(10)
    integer                       :: temp
    integer,          allocatable :: group_list(:), atom_no(:)
    real(wp)                      :: mm_mass
    type(s_selatoms), allocatable :: selatoms(:)
    integer                       :: kalloc_stat, kdealloc_stat
    logical                       :: found_atom, setup_qmmm_error

    return

  end subroutine setup_qmmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mmlist
  !> @brief        setup for MM list
  !> @authors      KY
  !! @param[in]    molecule  : molecular information
  !! @param[out]   qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mmlist(molecule, qmmm)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule
    type(s_qmmm),            intent(inout) :: qmmm

    ! local variables
    integer :: i, id, nres, natm, max_natm, res_no1, res_no2

    return

  end subroutine setup_mmlist

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_check_input_qchem
  !> @brief        check template for qchem input
  !! @authors      KYMD
  !! @param[in]    qmmm        : information on QM/MM setting
  !! @param[inout] setup_error : return true, when error is detected
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine qm_check_input_qchem(qmmm, setup_error)

    ! formal arguments
    type(s_qmmm), intent(inout) :: qmmm
    logical, intent(inout)      :: setup_error


    return

  end subroutine qm_check_input_qchem

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_check_input_gaussian
  !> @brief        check template for gaussian input
  !! @authors      KY
  !! @param[in]    qmmm        : information on QM/MM setting
  !! @param[inout] setup_error : return true, when error is detected
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine qm_check_input_gaussian(qmmm, setup_error)

    ! formal arguments
    type(s_qmmm), intent(inout) :: qmmm
    logical, intent(inout)      :: setup_error

    ! local variables
    logical                     :: defined_coord, defined_charge
    integer                     :: file1, ierror
    character(256)              :: string

    return

  end subroutine qm_check_input_gaussian

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_check_input_gaussian_frwf
  !> @brief        check template for gaussian input
  !! @authors      KYMD, KY
  !! @param[in]    qmmm        : information on QM/MM setting
  !! @param[inout] setup_error : return true, when error is detected
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine qm_check_input_gaussian_frwf(qmmm, setup_error)

    ! formal arguments
    type(s_qmmm), intent(inout) :: qmmm
    logical, intent(inout)      :: setup_error

    ! local variables
    logical                     :: is_route_section
    logical                     :: defined_gen, defined_iop6, defined_iop4,  &
                                   defined_coord, defined_charge, defined_elfield
    integer                     :: file1, ipos, ierror, is_exist
    character(256)              :: string

    return

  end subroutine qm_check_input_gaussian_frwf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_check_input_molpro
  !> @brief        check template for molpro input
  !! @authors      YA
  !! @param[in]    qmmm        : information on QM/MM setting
  !! @param[inout] setup_error : return true, when error is detected
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine qm_check_input_molpro(qmmm, setup_error)

    ! formal arguments
    type(s_qmmm), intent(inout) :: qmmm
    logical, intent(inout)      :: setup_error


    return

  end subroutine qm_check_input_molpro

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_atomic_number
  !> @brief        get atomic number
  !! @authors      YA, KY
  !! @param[in]    mm_mass   : atom mass
  !! @param[in]    atomic_no : atomic number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_atomic_number(mm_mass, atomic_no)

    ! formal arguments
    real(wp),                intent(in)  :: mm_mass
    integer                              :: atomic_no

    ! local variables
    integer                  :: ielem


    return

  end subroutine qm_atomic_number

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_generate_input
  !> @brief        generate QM input file
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of whole molecule
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_generate_input(molecule, coord, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm),            intent(inout) :: qmmm

    ! local variables
    character(4)  :: string
    integer       :: i, qm_atom, mm_atom
    real(wp)      :: rr01, r0(3), r1(3), r01(3)


    return

  end subroutine qm_generate_input

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_read_output
  !> @brief        Retrieve energy and gradients from QM output files
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of molecules
  !! @param[in]    qmmm      : QM/MM information
  !! @param[inout] energy    : energy information
  !! @param[inout] energy    : atomic forces
  !! @param[inout] qm_error  : error flag
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_read_output(molecule, coord, qmmm, energy, force, qm_error)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm), target,    intent(inout) :: qmmm
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:)
    logical,                 intent(inout) :: qm_error


    ! local variables
    real(wp), pointer :: qm_force(:,:), mm_force(:,:), la_force(:,:)
    real(wp), pointer :: qm_charge(:), la_charge(:)
    real(wp), pointer :: qm_dipole(:)
    real(wp)          :: qm_energy
    real(wp)          :: diff

    integer           :: i, j, n, address, qm_atom, mm_atom
    integer           :: errorcode
    real(wp)          :: rab, vab(3), rab1, rab3, ff, lafn(3), la_chg
    character(256)    :: exec


    return

  end subroutine qm_read_output

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_qminfo
  !> @brief        write information of QM calc. to file
  !! @authors      KY
  !! @param[in]    qmmm
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine write_qminfo(qmmm, qm_energy)

    ! formal arguments
    type(s_qmmm),            intent(in) :: qmmm
    real(wp),                intent(in) :: qm_energy

    ! local variables
    character(256)           :: qminfo
    integer                  :: file1, i

    return

  end subroutine write_qminfo

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_monitor
  !> @brief        Monitor QM execution
  !! @authors      YA
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_monitor(qmmm)

    ! formal arguments
    type(s_qmmm),            intent(in) :: qmmm

    ! local variables
    integer,    parameter :: monitor_interval = 1
    integer,    parameter :: monitor_max = 3600
    character(128)        :: filename, exec
    logical               :: done
    integer               :: i

    return

  end subroutine qm_monitor

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_mmlist
  !> @brief        Update list of MM atoms that are used for QM calc.
  !! @authors      KY
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_mmlist(coord, qmmm)

    ! formal arguments
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm), target,    intent(inout) :: qmmm

    ! local variables
    integer :: n, i, j, imm, jqm, ii
    integer, pointer :: mmatom_id(:), qmatom_id(:)
    integer, pointer :: mm_natoms_res(:), mmatom_id_res(:,:)
    real(wp)         :: cutoffdist2, dij(3), rij2
    integer :: id, my_id, nthread, nsize
#ifdef OMP
    integer :: omp_get_thread_num, omp_get_max_threads
#endif
    integer, allocatable :: mm_natoms(:), mmatom_id_n(:,:)
    logical              :: add_res
    logical              :: byresidue

    return

  end subroutine update_mmlist

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_generate_input_qchem
  !> @brief        generate QChem input file
  !! @authors      YA
  !! @param[in]    molecule  : information of whole molecule
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_generate_input_qchem(molecule, coord, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm),            intent(in)    :: qmmm

    ! local variables
    character(256)           :: qminp
    character(1024)          :: string, string2
    character(1)             :: conv
    integer                  :: file1, file2
    integer                  :: i
    integer                  :: address, atomic_no, charge, spin

    return

  end subroutine qm_generate_input_qchem

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_generate_input_gaussian
  !> @brief        generate Gaussian input file
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of whole molecule
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_generate_input_gaussian(molecule, coord, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm),            intent(in)    :: qmmm

    ! local variables
    character(256)           :: qminp
    character(1024)          :: string1, string2
    character(1)             :: conv
    integer                  :: file1, file2
    integer                  :: i
    integer                  :: address, atomic_no, charge, spin


    return

  end subroutine qm_generate_input_gaussian

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_generate_input_gaussian_frwf
  !> @brief        generate Gaussian input file
  !! @authors      KYMD, KY
  !! @param[in]    molecule  : information of whole molecule
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_generate_input_gaussian_frwf(molecule, coord, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm),            intent(in)    :: qmmm

    ! local variables
    character(256)           :: qminp
    character(256)           :: string, string_org
    integer                  :: file1, file2
    integer                  :: i ! , ii
    integer                  :: address, atomic_no, charge, spin

    logical                  :: defined_gen
    integer                  :: ierror

    return

  end subroutine qm_generate_input_gaussian_frwf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_generate_input_molpro
  !> @brief        generate Molpro input file
  !! @authors      YA
  !! @param[in]    molecule  : information of whole molecule
  !! @param[in]    coord     : atomic coordinates
  !! @param[in]    qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_generate_input_molpro(molecule, coord, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    real(wp),                intent(in)    :: coord(:,:)
    type(s_qmmm),            intent(in)    :: qmmm

    ! local variables
    character(256)           :: qminp, qminppc
    character(256)           :: string, string2
    integer                  :: file1, file2, file3
    integer                  :: i
    integer                  :: address, atomic_no, charge, spin

    return

  end subroutine qm_generate_input_molpro

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_read_output_qchem
  !> @brief        retrieve energy and gradients from QChem output files
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of molecules
  !! @param[in]    qmmm      : QM/MM information
  !! @param[inout] qm_energy : QM energy
  !! @param[inout] qm_error  : error flag
  !! @note  : This routine is called only from replica_main_rank
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_read_output_qchem(molecule, qmmm, qm_energy, qm_error)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm), target,    intent(inout) :: qmmm
    real(wp),                intent(inout) :: qm_energy
    logical,                 intent(inout) :: qm_error

    ! local variables
    character(256)           :: line, efield, exec, fname
    logical                  :: exist_energy, exist_fchk, exist_force, ex
    logical                  :: exist_fchk42
    logical                  :: found_energy, found_dipole, found_force
    integer                  :: file1, fchk
    integer                  :: i
    integer                  :: nsta, nend
    integer                  :: address, atomic_no, charge, spin
    real(wp)                 :: total_energy, mm_energy, force_tmp(3)

    real(wp), pointer :: qm_force(:,:), mm_force(:,:), la_force(:,:)

    real(wp), parameter  :: conv_debye_au = 3.934303E-01

    return

  end subroutine qm_read_output_qchem

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_read_output_gaussian
  !> @brief        retrieve energy and gradients from Gaussian output files
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of molecules
  !! @param[in]    qmmm      : QM/MM information
  !! @param[inout] qm_energy : QM energy
  !! @param[inout] qm_error  : error flag
  !! @note  : This routine is called only from replica_main_rank
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_read_output_gaussian(molecule, qmmm, qm_energy, qm_error)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm), target,    intent(inout) :: qmmm
    real(wp),                intent(inout) :: qm_energy
    logical,                 intent(inout) :: qm_error

    ! local variables
    character(256)       :: line, fchk, exec
    logical              :: exist_log, exist_fchk
    logical              :: found_energy_total, found_energy_charge
    logical              :: found_force, found_field, found_charge
    integer              :: file1, file2
    integer              :: i, j
    integer              :: nsta, nend
    integer              :: address
    integer              :: ndata, nlines
    integer              :: idum
    real(wp)             :: rdum
    real(wp)             :: total_energy, mm_energy
    real(wp)             :: force_tmp((qmmm%qm_natoms+qmmm%num_qmmmbonds)*3)
    real(wp)             :: charge_tmp(qmmm%qm_natoms+qmmm%num_qmmmbonds) 

    real(wp), pointer    :: qm_force(:,:), mm_force(:,:), la_force(:,:)
    real(wp), pointer    :: qm_charge(:), la_charge(:)

    return

  end subroutine qm_read_output_gaussian

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_read_output_gaussian_frwf
  !> @brief        retrieve energy and gradients from Gaussian FORMATTED rwf,
  !                which is generated with system-called formrwf.
  !                Please contact KYMD to have formrwf program. 
  !! @authors      YA, KY, KYMD
  !! @param[in]    molecule  : information of molecules
  !! @param[in]    qmmm      : QM/MM information
  !! @param[inout] qm_energy : QM energy
  !! @param[inout] qm_error  : error flag
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_read_output_gaussian_frwf(molecule, qmmm, qm_energy, qm_error)

    ! formal arguments
    type(s_molecule),     intent(in)    :: molecule
    type(s_qmmm), target, intent(inout) :: qmmm
    real(wp),             intent(inout) :: qm_energy
    logical,              intent(inout) :: qm_error

    ! local variables
    character(256)      :: line, rwf, frwf, exec
    logical             :: exist_log, exist_rwf, exist_frwf
    logical             :: found_energy, found_force, found_charge, &
                           found_dipole, found_selfe, found_efield
    integer             :: file1
    integer             :: i, j
    integer             :: nsta, nend
    integer             :: address
    integer             :: ndata, nlines
    integer             :: idum
    real(wp)            :: rdum
    real(wp)            :: total_energy, mm_energy
    real(wp)            :: force_tmp((qmmm%qm_natoms+qmmm%num_qmmmbonds)*3)
    real(wp)            :: charge_tmp(qmmm%qm_natoms+qmmm%num_qmmmbonds) 

    real(wp), pointer   :: qm_force(:,:), mm_force(:,:), la_force(:,:), qm_charge(:), la_charge(:)

    real(wp), parameter :: conv_debye_au = 3.934303E-01

    integer             :: ierror, natom_qmlink, jatom_mm
    real(wp)            :: field_tmp(1: 5)

    return

  contains

    !====1=========2=========3=========4=========5=========6=========7=========8
    !
    !  Subroutine    pass_value
    !> @brief        pass value in 1D vector to 2D matrix
    !! @authors      KYMD
    !! @param[in]    natom_mm   : # of MM atoms
    !! @param[in]    istart     : # of first item in read line
    !! @param[in]    iend       : # of last item in read line
    !! @param[in]    jpos_save  : Position of column in 2D matrix
    !! @param[in]    field_tmp  : 1D vector
    !! @param[out]   mm_force   : 2D matrix
    !
    !====1=========2=========3=========4=========5=========6=========7=========8

    subroutine pass_value(natom_mm, istart, iend, jpos_save, field_tmp, mm_force)

      ! formal arguments
      integer, intent(in)    :: natom_mm, istart, iend
      integer                :: jpos_save
      real(wp), intent(in)   :: field_tmp(5)
      real(wp)               :: mm_force(3, natom_mm)

      ! local variables
      integer :: ipass_pattern, iend_pattern
      integer :: i, j, jend

      return
    end subroutine pass_value

    !====1=========2=========3=========4=========5=========6=========7=========8
    !
    !  Subroutine    pass_value2
    !> @brief        pass value in 1D vector to 2D matrix
    !! @authors      KYMD
    !! @param[in]    natom_mm   : # of MM atoms
    !! @param[in]    istart     : # of first item in read line
    !! @param[in]    iend       : # of last item in read line
    !! @param[in]    field_tmp  : 1D vector
    !! @param[out]   mm_force   : 2D matrix
    !
    !====1=========2=========3=========4=========5=========6=========7=========8

    subroutine pass_value2(natom_mm, istart, iend, field_tmp, mm_force)

      ! formal arguments
      integer, intent(in)    :: natom_mm, istart, iend
      real(wp), intent(in)   :: field_tmp(5)
      real(wp)               :: mm_force(1: 3*natom_mm, 1: 1)

      ! local variables
      integer :: jend

      return
    end subroutine pass_value2

  end subroutine qm_read_output_gaussian_frwf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    qm_read_output_molpro
  !> @brief        retrieve energy and gradients from Molpro output files
  !! @authors      YA, KY
  !! @param[in]    molecule  : information of molecules
  !! @param[in]    qmmm      : QM/MM information
  !! @param[inout] qm_energy : QM energy
  !! @param[inout] qm_error  : error flag
  !! @note  : This routine is called only from replica_main_rank
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine qm_read_output_molpro(molecule, qmmm, qm_energy, qm_error)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm), target,    intent(inout) :: qmmm
    real(wp),                intent(inout) :: qm_energy
    logical,                 intent(inout) :: qm_error

    ! local variables
    character(256)       :: line
    logical              :: exist_log
    logical              :: found_energy, found_force, found_mmforce
    logical              :: found_dipole
    integer              :: file1
    integer              :: i, j
    integer              :: nsta, nend
    integer              :: idum
    real(wp)             :: total_energy, mm_energy
    real(wp)             :: force_tmp((qmmm%qm_natoms+qmmm%num_qmmmbonds)*3)

    real(wp), pointer    :: qm_force(:,:), mm_force(:,:), la_force(:,:)
    real(wp), pointer    :: qm_charge(:), la_charge(:)

    return

  end subroutine qm_read_output_molpro

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_qmmmbonds
  !> @brief        get number of QM-MM bonds
  !! @authors      YA
  !! @param[in]    molecule  : molecular information
  !! @param[inout] qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine count_qmmmbonds(molecule, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm),            intent(inout) :: qmmm

    ! local variables
    integer                  :: list_bond(10), list_bond_0(10)
    integer                  :: num_bond
    integer                  :: i, j, k, l, iatom, ibond

    return

  end subroutine count_qmmmbonds

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    generate_linkatoms
  !> @brief        generate link atom bond_list and exclude MM atoms
  !! @authors      YA, KY
  !! @param[in]    molecule  : molecular information
  !! @param[in]    psf       : PSF information
  !! @param[inout] qmmm      : QM/MM information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine generate_linkatoms(molecule, psf, qmmm)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_psf),             intent(in)    :: psf
    type(s_qmmm),            intent(inout) :: qmmm

    ! local variables
    integer                  :: list_bond(10), list_bond_0(10)
    integer                  :: num_bond, num_linkatoms
    integer                  :: i, j, k, l, ista, iend, iatom, ibond, igrp, natom_grp, iatom_rm
    integer                  :: lkatm
    integer                  :: locsta, locend
    real(wp)                 :: rr01, r0(3), r1(3), r01(3)

    integer  :: max_ecatoms, num_ecatoms, nec0
    integer  :: kalloc_stat
    integer, parameter :: num_ecatm_per_link = 20

    return

  end subroutine generate_linkatoms

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_linkmm
  !> @brief        check that link MM atom is NOT hydrogen 
  !! @authors      KY
  !! @param[in]    molecule    : molecular information
  !! @param[inout] qmmm        : QM/MM information
  !! @param[inout] setup_error : true, when error is detected
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine check_linkmm(molecule, qmmm, setup_error)

    ! formal arguments
    type(s_molecule), intent(in)    :: molecule
    type(s_qmmm),     intent(inout) :: qmmm
    logical,          intent(inout) :: setup_error

    ! local variables
    integer  :: i, lkmm, ano
    real(wp) :: mm_mass
    logical  :: is_hydrogen

  end subroutine check_linkmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    rearrange_term_list
  !> @brief        rearrange the list of terms in such a way that 
  !!                     list(:, 1:mm_terms)                   ... MM terms
  !!                     list(:, mm_terms:mm_terms + qm_terms) ... QM terms
  !! @authors      KY
  !! @param[in]    ityp      : type of force field terms
  !!                           2 = bond, 3 = angle, 4 = dihed, CMAP
  !!                           5 = imprpr
  !! @param[in]    qm_natoms : number of QM atoms
  !! @param[in]    qmatom_id : ID of QM atoms
  !! @param[out]   qm_terms  : number of terms in QM region
  !! @param[in]    num_terms : number of all terms 
  !! @param[out]   list      : list of terms
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine rearrange_term_list(ityp, qm_natoms, qmatom_id, &
                                    qm_terms, num_terms, list)

    integer, intent(in)    :: ityp
    integer, intent(in)    :: qm_natoms
    integer, intent(in)    :: qmatom_id(qm_natoms)
    integer                :: qm_terms
    integer, intent(in)    :: num_terms
    integer                :: list(ityp,num_terms)

    integer, allocatable :: qm_index(:)
    integer, allocatable :: mm_index(:)
    integer, allocatable :: qm_list (:,:)

    integer  :: i, n, mm_terms
    integer  :: j, jj, jmax
    integer, allocatable :: jlist(:)
    integer              :: kalloc_stat, kdealloc_stat

    return

  end subroutine rearrange_term_list

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    rearrange_impr_list
  !> @brief        rearrange the list of terms in such a way that 
  !!                     list(:, 1:mm_terms)                   ... MM terms
  !!                     list(:, mm_terms:mm_terms + qm_terms) ... QM terms
  !! @authors      KY
  !! @param[in]    qm_natoms : number of QM atoms
  !! @param[in]    qmatom_id : ID of QM atoms
  !! @param[out]   qm_terms  : number of terms in QM region
  !! @param[in]    num_terms : number of all terms 
  !! @param[out]   list      : list of terms
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine rearrange_impr_list(qm_natoms, qmatom_id, &
                                 qm_terms, num_terms, list)

    integer, intent(in)    :: qm_natoms
    integer, intent(in)    :: qmatom_id(qm_natoms)
    integer                :: qm_terms
    integer, intent(in)    :: num_terms
    integer                :: list(4,num_terms)

    integer, allocatable :: qm_index(:)
    integer, allocatable :: mm_index(:)
    integer, allocatable :: qm_list (:,:)

    integer  :: i, n, mm_terms
    integer  :: j, jj, jmax
    !integer, allocatable :: jlist(:)
    integer              :: kalloc_stat, kdealloc_stat

    return

  end subroutine rearrange_impr_list

end module at_qmmm_mod
