!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_rpath_mep_mod
!> @brief   Minimum energy path search using replicas
!! @authors Yoshinobu Akinaga (YA), Kiyoshi Yagi (KY)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_rpath_mep_mod

  use at_energy_mod
  use at_dynamics_str_mod
  use at_dynvars_str_mod
  use at_dynvars_mod
  use at_rpath_str_mod
  use at_boundary_str_mod
  use at_boundary_mod
  use at_pairlist_str_mod
  use at_pairlist_mod
  use at_enefunc_str_mod
  use at_minimize_str_mod
  use at_minimize_mod
  use at_output_str_mod
  use at_output_mod
  use molecules_str_mod
  use fileio_rst_mod
  use fileio_rstmep_mod
  use select_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use math_libs_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: setup_rpath_mep
  public  :: setup_mepatoms
  public  :: setup_mepatoms_qmmm
  public  :: run_rpath_mep
  private :: evolve_mep
  private :: reparametrize_mep
  private :: interpolation
  private :: energy_and_force
  private :: check_convergence
  private :: trans_mass_weight_coord
  private :: backtrans_mass_weight_coord

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_rpath_mep
  !> @brief        setup RPATH
  !! @authors      YA, KY
  !! @param[in]    mepatm_select_index : index of MEP atoms
  !! @param[in]    sel_info   : selector input information
  !! @param[in]    rst        : restart information
  !! @param[in]    rstmep     : rstmep data
  !! @param[in]    molecule   : molecule information
  !! @param[in]    qmmm       : qmmm information
  !! @param[inout] minimize   : minimize information
  !! @param[inout] dynvars    : dynamic variables
  !! @param[inout] rpath      : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_rpath_mep(mepatm_select_index, sel_info, rst, rstmep, &
                             molecule, qmmm, minimize, dynvars, rpath)

    ! formal arguments
    character(256),          intent(in)    :: mepatm_select_index
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_rst),             intent(in)    :: rst
    type(s_rstmep),          intent(in)    :: rstmep
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm),            intent(inout) :: qmmm
    type(s_minimize),        intent(inout) :: minimize
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_rpath),           intent(inout) :: rpath

    ! local variables
    integer   :: i, ii
    integer   :: replicaid
    integer   :: atomid, iatom
    integer   :: errorcode, ierr
    logical   :: err


    return

  end subroutine setup_rpath_mep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mepatoms
  !> @brief        define MEP atoms
  !! @authors      YA
  !! @param[in]    mepatm_select_index : index of MEP atoms
  !! @param[in]    sel_info   : selector input information
  !! @param[in]    molecule   : molecule information
  !! @param[out]   rpath      : rpath parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mepatoms(mepatm_select_index, sel_info, molecule, rpath)

    ! formal arguments
    character(256),      intent(in)    :: mepatm_select_index
    type(s_sel_info),    intent(in)    :: sel_info
    type(s_molecule),    intent(in)    :: molecule
    type(s_rpath),       intent(inout) :: rpath

    ! local variables
    integer                :: igroup, ngroup, natom, i, j, offset, temp
    integer, allocatable   :: group_list(:)
    type(s_selatoms), allocatable :: selatoms(:)


    return

  end subroutine setup_mepatoms

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mepatoms_qmmm
  !> @brief        define MEP atoms for QM/MM jobs
  !! @authors      KY
  !! @param[in]    molecule   : molecule information
  !! @param[in]    qmmm       : QMMM information
  !! @param[out]   rpath      : rpath parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mepatoms_qmmm(molecule, qmmm, rpath)

    ! formal arguments
    type(s_molecule),  intent(in)    :: molecule
    type(s_qmmm),      intent(in)    :: qmmm
    type(s_rpath),     intent(inout) :: rpath

    ! local variables
    integer              :: i, j, k, id, nad, ntmp
    logical              :: error
    logical, allocatable :: found(:)
    integer, allocatable :: add(:), tmp(:)

    return

  end subroutine setup_mepatoms_qmmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_rpath_mep
  !> @brief        control string method
  !! @authors      TM, YK, YA, KY
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] minimize    : minimize information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] rpath       : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_rpath_mep(output, molecule, enefunc, dynvars, minimize, &
                           dynamics, pairlist, boundary, rpath)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule),         intent(inout) :: molecule
    type(s_enefunc),  target, intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_minimize), target, intent(inout) :: minimize
    type(s_dynamics),         intent(inout) :: dynamics
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary
    type(s_rpath),            intent(inout) :: rpath

    ! local variables
    logical                :: conv
    integer                :: i, j, k, n
    integer                :: niter
    integer                :: iloop_start, iloop_end
    integer                :: replicaid
    integer                :: ii, jj, iatom, id

    integer                :: natom_micro
    integer, pointer       :: optatom_micro_id(:)
    real(wp)               :: energy0corr, e0
    real(wp), allocatable  :: coord0(:,:), force0corr(:,:)

    real(wp)    , pointer  :: coord(:,:), force(:,:)
    type(s_qmmm), pointer  :: qmmm

    character(256)         :: folder, basename
    character(5)           :: num
    logical                :: savefile

    integer                :: repid

    return

  end subroutine run_rpath_mep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    evolve_mep
  !> @brief        evolve path (MEP version)
  !! @authors      YK, YM, YA, KY
  !! @param[inout] dynvars  : dynamical variables information
  !! @param[inout] rpath    : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine evolve_mep(rpath)

    ! formal arguments
    type(s_rpath), target, intent(in) :: rpath

    ! local variables
    integer           :: dimension
    real(wp)          :: delta
    real(wp), pointer :: force(:)
    real(wp), pointer :: image(:,:)
    integer           :: i, repid


    return

  end subroutine evolve_mep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reparametrize_mep
  !> @brief        reparametrize path
  !! @authors      YA, KY
  !! @param[inout] rpath    : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reparametrize_mep(rpath)

    ! formal arguments
    type(s_rpath),    target, intent(inout) :: rpath

    ! local variables
    integer                  :: i, j, k
    integer                  :: dimno, dimno_i, dimno_j
    integer                  :: repid, repid_i, repid_j

    real(wp),    allocatable :: path(:,:), path_reparm(:,:)
    real(wp),    allocatable :: path_leng(:), path_equi(:)
    real(wp),    allocatable :: nonuniform_mesh(:)
    real(wp),    allocatable :: Vs(:)

    real(dp),        pointer :: before_gather(:), after_gather(:)
    real(wp),    allocatable :: after(:)

    return

  end subroutine reparametrize_mep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    interpolation
  !> @brief        update images
  !! @authors      YA
  !! @param[in]    nn           : # of points
  !! @param[in]    path         : images
  !! @param[in]    qi           : non-uniform mesh
  !! @param[inout] Vs           : work space
  !! @param[out]   path_reparam : updated images
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine interpolation(nimages, qi, path, Vs, path_reparm)

    ! formal arguments
    integer,         intent(in) :: nimages
    real(wp),        intent(in) :: path(nimages)
    real(wp),        intent(in) :: qi(nimages)
    real(wp),        intent(in) :: Vs(nimages)
    real(wp),     intent(inout) :: path_reparm(nimages)

    ! local variables
    integer  :: i
    real(wp) :: tmp
    real(wp) :: coeff(4,nimages-1)


    return

  end subroutine interpolation

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    energy_and_force
  !> @brief        compute energy and force for each image
  !! @authors      YA
  !! @param[in]    rpath        : rpath info
  !! @param[in]    molecule     : molecular info
  !! @param[in]    enefunc      : enefunc info
  !! @param[in]    pairlist     : pairlist info
  !! @param[in]    boundary     : boundary info
  !! @param[in]    nonb_ene     : flag for calculate nonbonded energy
  !! @param[inout] dynvars      : dynvars info
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine energy_and_force(ncount, rpath, molecule, enefunc, pairlist, &
                              boundary, nonb_ene, dynvars)

    ! formal arguments
    integer,             intent(in) :: ncount
    type(s_rpath),    intent(inout) :: rpath
    type(s_molecule),    intent(inout) :: molecule  !YA
    type(s_enefunc),  intent(inout) :: enefunc
    type(s_pairlist),    intent(in) :: pairlist
    type(s_boundary),    intent(in) :: boundary
    logical,             intent(in) :: nonb_ene
    type(s_dynvars),  intent(inout) :: dynvars

    ! local variables
    character(256)                  :: folder, basename
    character(5)                    :: num
    logical                         :: savefile
    integer                         :: repid
    integer                         :: dimno, iatom, i
    real(wp)                        :: x, y, x2, y2, r

    return

  end subroutine energy_and_force

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_convergence
  !> @brief        
  !! @authors      YA
  !! @param[in]    rpath        : rpath info
  !! @param[in]    molecule     : molecular info
  !! @param[in]    enefunc      : enefunc info
  !! @param[in]    pairlist     : pairlist info
  !! @param[in]    boundary     : boundary info
  !! @param[in]    nonb_ene     : flag for calculate nonbonded energy
  !! @param[inout] dynvars      : dynvars info
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_convergence(rpath, niter, conv)

    ! formal arguments
    type(s_rpath),    intent(inout) :: rpath
    integer,             intent(in) :: niter
    logical                         :: conv

    ! local variables
    logical                         :: my_conv, conv_e, conv_path
    integer                         :: repid, i
    real(wp)                        :: disp, delta, dmax, sum
    real(wp),           allocatable :: work(:), work1(:)


    return

  end subroutine check_convergence

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    trans_mass_weight_coord
  !> @brief        
  !! @authors      YA, KY
  !! @param[in]    molecule     : molecular info
  !! @param[inout] rpath        : rpath info
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine trans_mass_weight_coord(molecule, rpath)

    ! formal arguments
    type(s_molecule),    intent(in) :: molecule
    type(s_rpath),    intent(inout) :: rpath

    ! local variables
    integer                         :: repid
    integer                         :: i, ii, iatom, offset
    real(wp)                        :: massfac

    return

  end subroutine trans_mass_weight_coord

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    backtrans_mass_weight_coord
  !> @brief        
  !! @authors      YA, KY
  !! @param[in]    molecule     : molecular info
  !! @param[inout] rpath        : rpath info
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine backtrans_mass_weight_coord(molecule, rpath)

    ! formal arguments
    type(s_molecule),    intent(in) :: molecule
    type(s_rpath),    intent(inout) :: rpath

    ! local variables
    integer                         :: repid
    integer                         :: i, ii, iatom, offset
    real(wp)                        :: massfac

    return

  end subroutine backtrans_mass_weight_coord


end module at_rpath_mep_mod
